#include <Wire.h>
#include <LiquidCrystal_I2C.h>

// LCD Configuration with custom characters
LiquidCrystal_I2C lcd(0x27, 16, 2);

// Custom character definitions
byte alarmChar[8] = {
  B00100,
  B01110,
  B01110,
  B01110,
  B11111,
  B00000,
  B00100,
  B00000
};

byte batteryLow[8] = {
  B01110,
  B11011,
  B10001,
  B10001,
  B10001,
  B10001,
  B11111,
  B11111
};

byte batteryMed[8] = {
  B01110,
  B11011,
  B10001,
  B10001,
  B11111,
  B11111,
  B11111,
  B11111
};

byte batteryFull[8] = {
  B01110,
  B11111,
  B11111,
  B11111,
  B11111,
  B11111,
  B11111,
  B11111
};

// Pin Definitions
#define ENC_CLK 3
#define ENC_DT 7
#define ENC_SW 4
#define BUZZER_PIN 13 
#define THERMISTOR_PIN A3
#define BATTERY_PIN A0

// Thermistor Configuration
#define SERIES_RESISTOR 100000    // 1.5kΩ resistor
#define THERMISTOR_NOMINAL 10000
#define TEMPERATURE_NOMINAL 25
#define B_COEFFICIENT 3950
#define NUM_SAMPLES 5

// Battery Configuration
#define BATTERY_MIN 3.0
#define BATTERY_MAX 4.2
#define BATTERY_OFFSET 150      // Calibration offset

// Temperature Unit Configuration
const bool useFahrenheit = true;

// Global Variables
float targetTempC = 100.0;
float currentTempC = 0.0;
int batteryPercent = 0;
volatile int encoderPos = 0;
int lastEncPos = 0;
bool alarmEnabled = true;

// Function Prototypes
void updateEncoder();
void handleEncoderInput();
void updateDisplay();
void printTemperature(float tempC);
void checkAlarm();
int readBattery();
int readADC();
float calculateResistance(int adcValue);
float calculateTemperature(float resistance);

// ================== SETUP ==================
void setup() {
  // Encoder setup
  pinMode(ENC_CLK, INPUT_PULLUP);
  pinMode(ENC_DT, INPUT_PULLUP);
  pinMode(ENC_SW, INPUT_PULLUP);
  attachInterrupt(digitalPinToInterrupt(ENC_CLK), updateEncoder, CHANGE);

  // Buzzer setup
  pinMode(BUZZER_PIN, OUTPUT);

  // LCD setup
  lcd.init();
  lcd.createChar(0, alarmChar);
  lcd.createChar(1, batteryLow);
  lcd.createChar(2, batteryMed);
  lcd.createChar(3, batteryFull);
  lcd.backlight();
  lcd.clear();
}

// ================== MAIN LOOP ==================
void loop() {
  // Read sensors with new formulas
  int adcValue = readADC();
  float resistance = calculateResistance(adcValue);
  currentTempC = calculateTemperature(resistance);
  
  batteryPercent = readBattery();
  
  handleEncoderInput();
  updateDisplay();
  checkAlarm();
  delay(100);
}

// ================== ENCODER FUNCTIONS ==================
void updateEncoder() {
  static int oldState = HIGH;
  int newState = digitalRead(ENC_CLK);
  if (newState != oldState) {
    if (digitalRead(ENC_DT) != newState) {
      encoderPos++;
    } else {
      encoderPos--;
    }
    oldState = newState;
  }
}

void handleEncoderInput() {
  // Handle temperature adjustment
  if (encoderPos != lastEncPos) {
    float change = (encoderPos > lastEncPos) ? 1.0 : -1.0;
    targetTempC = constrain(targetTempC + change, -20.0, 120.0); // Range: -20°C to 120°C
    lastEncPos = encoderPos;
  }

  // Handle button press
  if (digitalRead(ENC_SW) == LOW) {
    delay(50);  // Debounce
    alarmEnabled = !alarmEnabled;
    while (digitalRead(ENC_SW) == LOW);  // Wait for button release
  }
}

// ================== DISPLAY FUNCTIONS ==================
void updateDisplay() {
  // Target temperature line
  lcd.setCursor(0, 0);
  lcd.print("tar:");
  printTemperature(targetTempC);

  // Battery icon and percentage
  lcd.setCursor(10, 0);
  if (batteryPercent <= 30) lcd.write(1);
  else if (batteryPercent <= 70) lcd.write(2);
  else lcd.write(3);
  lcd.print(":");
  lcd.print(batteryPercent);
  lcd.print("%");

  // Current temperature line
  lcd.setCursor(0, 1);
  lcd.print("cur:");
  printTemperature(currentTempC);

  // Alarm symbol and text
  lcd.setCursor(10, 1); // Position (10, 1)
  alarmEnabled ? lcd.write(0) : lcd.print(" ");
  lcd.print("alarm"); // Add "alarm" text after the symbol
}

void printTemperature(float tempC) {
  int temp = useFahrenheit ? (tempC * 9 / 5) + 32 : tempC;
  if (temp < 100) lcd.print(" ");
  if (temp < 10) lcd.print(" ");
  lcd.print(temp);
  lcd.print(useFahrenheit ? "F" : "C");
}

// ================== BUZZER FUNCTIONS ==================
void checkAlarm() {
  if (alarmEnabled && (currentTempC >= targetTempC)) {
    tone(BUZZER_PIN, 1000); // Play tone at 1000Hz
    delay(200);
    noTone(BUZZER_PIN);
    delay(200);
  } else {
    noTone(BUZZER_PIN); // Turn off buzzer
  }
}

// ================== TEMPERATURE CALCULATION ==================
int readADC() {
  int adcValue = 0;
  for (int i = 0; i < NUM_SAMPLES; i++) {
    adcValue += analogRead(THERMISTOR_PIN);
    delay(10);
  }
  return adcValue / NUM_SAMPLES;
}

float calculateResistance(int adcValue) {
  if (adcValue == 0) return INFINITY; // Prevent division by zero
  float voltage = adcValue * (5.0 / 1023.0);
  return SERIES_RESISTOR * ((5.0 / voltage) - 1);
}

float calculateTemperature(float resistance) {
  float steinhart = log(resistance / THERMISTOR_NOMINAL);
  steinhart /= B_COEFFICIENT;
  steinhart += 1.0 / (TEMPERATURE_NOMINAL + 273.15);
  return (1.0 / steinhart) - 273.15;
}

// ================== BATTERY CALCULATION ==================
int readBattery() {
  int adcValue = analogRead(BATTERY_PIN) - BATTERY_OFFSET;
  adcValue = max(adcValue, 0);
  float voltage = adcValue * (5.0 / 1023.0);
  voltage = constrain(voltage, BATTERY_MIN, BATTERY_MAX);
  return map(voltage * 100, BATTERY_MIN * 100, BATTERY_MAX * 100, 0, 100);
}